from __future__ import annotations

from collections.abc import Iterable, Mapping
import dataclasses
import math
from typing import Any, Self

import pandas as pd

from ..common import ExtendedFloat, encode_extended_float


type KeyItem = float | int | str


type Key = tuple[KeyItem, ...] | KeyItem


type Value = float | int


def is_value(arg: Any) -> bool:
    return isinstance(arg, float | int)


type DimensionArgument = Iterable[KeyItem]


type SparseTensorArgument = (
    Mapping[Key, Value]
    | pd.Series
    | pd.DataFrame  # For indicator parameters
    | Iterable[Key]  # For indicator parameters
)


#: Input tensor creation argument
type TensorArgument = (
    Value | SparseTensorArgument | tuple[SparseTensorArgument, Value]
)


@dataclasses.dataclass
class Tensor:
    """An n-dimensional matrix"""

    entries: list[Any]
    """Raw list of matrix entries"""

    default_value: ExtendedFloat = 0
    """Value to use for missing key"""

    @classmethod
    def from_argument(
        cls,
        arg: TensorArgument,
        rank: int,
        is_indicator: bool = False,
        is_pin: bool = False,
    ) -> Self:
        """Creates a tensor from a variety of argument values

        In most cases you will not need to call this method directly: it is
        called automatically during parameter validation.

        Args:
            arg: Raw argument to be wrapped into a tensor. This argument must
                match the tensor's shape, see below for details.
            rank: The expected rank of the tensor
            is_indicator: Whether the tensor holds indicator values
            is_pin: Whether the tensor corresponds to a pinned variable's
                parameter. This changes the default value from zero to minus
                infinity (unpinned).

        The accepted arguments depend on the tensor's domain rank (the length
        of its keys) and image (does it hold arbitrary numbers or only 0s/1s).
        If the tensor's rank is 0, i.e. is a scalar, the only accepted argument
        type is a number matching the image. Otherwise the following inputs are
        accepted:

        + Mapping (e.g. Python dictionary) with tuple key if rank is greater
          than 1
        + `pandas` series with key index

        Additionally, if the tensor holds indicator values, two more
        conveniences are allowed, each representing the set of keys which have
        value 1:

        + Iterable of keys
        + `pandas` dataframe

        Finally, all non-scalar tensor arguments can be wrapped into a tuple
        `(arg, default)` to provide a `default` value to use when no matching
        key exists.
        """
        if isinstance(arg, tuple):
            data, default_value = arg
        elif rank > 0 and is_value(arg) and not is_indicator:
            data = {}
            default_value = arg
        else:
            data = arg
            default_value = -math.inf if is_pin else 0
        if (
            is_indicator
            and isinstance(data, pd.Series)
            and not pd.api.types.is_numeric_dtype(data)
        ):
            data = data.reset_index()
        keyifier = _Keyifier(rank)
        if is_indicator and isinstance(data, pd.DataFrame):
            entries = [
                {"key": keyifier.keyify(key), "value": 1}
                for key in data.itertuples(index=False, name=None)
            ]
        elif is_indicator and not hasattr(data, "items"):
            entries = [
                {"key": keyifier.keyify(key), "value": 1} for key in data
            ]
        elif is_value(data):
            entries = [{"key": (), "value": encode_extended_float(data)}]
        else:
            entries = [
                {
                    "key": keyifier.keyify(key),
                    "value": encode_extended_float(value),
                }
                for key, value in data.items()
            ]
        return cls(entries, encode_extended_float(default_value))


class _Keyifier:
    def __init__(self, rank: int) -> None:
        self.rank = rank

    def keyify(self, key: KeyItem | Key) -> Key:
        tup = tuple(key) if isinstance(key, list | tuple) else (key,)
        if len(tup) != self.rank:
            raise Exception(f"Invalid key rank: {len(tup)} != {self.rank}")
        return tup



from collections.abc import AsyncIterator
import contextlib
import logging

import aiohttp
import brotli  # type: ignore

from .common import (
    CONTENT_TYPE_HEADER,
    TRACE_HEADER,
    BinaryExecutorResult,
    Executor,
    ExecutorError,
    ExecutorResult,
    Headers,
    JsonExecutorResult,
    JsonSeqExecutorResult,
    PlainTextExecutorResult,
)


_logger = logging.getLogger(__name__)


_BROTLI_QUALITY = 4


_COMPRESSION_THRESHOLD = 2**16


_READ_BUFFER_SIZE = 2**26  # 512MiB


_REQUEST_TIMEOUT_SECONDS = 900  # 15 minutes


class AiohttpExecutor(Executor):
    """`aiohttp`-powered executor"""

    def __init__(
        self, endpoint: str, authorization: str | None = None
    ) -> None:
        super().__init__(
            variant="aiohttp",
            endpoint=endpoint,
            authorization=authorization,
            supports_streaming=True,
        )

    @contextlib.asynccontextmanager
    async def _send(
        self, url: str, method: str, headers: Headers, body: bytes | None
    ) -> AsyncIterator[ExecutorResult]:
        if body and len(body) > _COMPRESSION_THRESHOLD:
            headers["content-encoding"] = "br"
            compressed_body = brotli.compress(
                body,
                mode=brotli.MODE_TEXT,
                quality=_BROTLI_QUALITY,
            )
            _logger.debug(
                "Compressed request body. [size=%s]", len(compressed_body)
            )
            body = compressed_body
        try:
            async with (
                aiohttp.ClientSession(
                    headers=headers,
                    read_bufsize=_READ_BUFFER_SIZE,
                    timeout=aiohttp.ClientTimeout(
                        total=_REQUEST_TIMEOUT_SECONDS
                    ),
                ) as session,
                session.request(url=url, method=method, data=body) as res,
            ):
                status = res.status
                trace = res.headers.get(TRACE_HEADER)
                ctype = res.headers.get(CONTENT_TYPE_HEADER)
                if JsonExecutorResult.is_eligible(ctype):
                    text = await res.text()
                    yield JsonExecutorResult(
                        status=status,
                        trace=trace,
                        text=text,
                    )
                elif JsonSeqExecutorResult.is_eligible(ctype):
                    yield JsonSeqExecutorResult(
                        status=status,
                        trace=trace,
                        reader=res.content,
                    )
                elif PlainTextExecutorResult.is_eligible(ctype):
                    yield PlainTextExecutorResult(
                        status=status,
                        trace=trace,
                        reader=res.content,
                    )
                elif BinaryExecutorResult.is_eligible(ctype):
                    yield BinaryExecutorResult(
                        status=status,
                        trace=trace,
                        reader=res.content,
                    )
                else:
                    text = await res.text()
                    raise ExecutorError(
                        status=status, trace=trace, reason=text
                    )
        except aiohttp.ClientResponseError as err:
            trace = None
            if isinstance(err.headers, list):
                trace = next(
                    (v for (k, v) in err.headers if k == TRACE_HEADER),
                    None,
                )
            raise ExecutorError(status=err.status, trace=trace) from err



from collections.abc import AsyncIterator
import contextlib
import logging

from pyodide.http import pyfetch  # type: ignore

from .common import (
    CONTENT_TYPE_HEADER,
    TRACE_HEADER,
    BinaryExecutorResult,
    Executor,
    ExecutorError,
    ExecutorResult,
    Headers,
    JsonExecutorResult,
    PlainTextExecutorResult,
)


_logger = logging.getLogger(__name__)


class PyodideExecutor(Executor):
    """`pyodide`-powered executor"""

    def __init__(
        self, endpoint: str, authorization: str | None = None
    ) -> None:
        super().__init__(
            variant="pyodide", endpoint=endpoint, authorization=authorization
        )

    @contextlib.asynccontextmanager
    async def _send(
        self, url: str, method: str, headers: Headers, body: bytes | None
    ) -> AsyncIterator[ExecutorResult]:
        res = await pyfetch(
            url=url,
            method=method,
            headers=headers,
            body=body,
        )
        status = res.status
        headers = res.js_response.headers
        trace = headers.get(TRACE_HEADER)
        ctype = headers.get(CONTENT_TYPE_HEADER)
        if JsonExecutorResult.is_eligible(ctype):
            text = await res.js_response.text()
            yield JsonExecutorResult(status=status, trace=trace, text=text)
        elif PlainTextExecutorResult.is_eligible(ctype):
            text = await res.js_response.text()
            yield PlainTextExecutorResult(
                status=status, trace=trace, reader=text
            )
        elif BinaryExecutorResult.is_eligible(ctype):
            data = await res.js_response.bytes()
            yield BinaryExecutorResult(status=status, trace=trace, reader=data)
        else:
            text = await res.js_response.text()
            raise ExecutorError(status=status, trace=trace, reason=text)
